#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#

import numpy as np
import pandas as pd
import os
import h5py
from pathlib import Path
from datetime import datetime
from .serdeutils import set_metadata_without_warning
from .utilities import ProjectException
from .utilities import TooManyBinomialValuesError
from .utilities import ValueConversionError
from .connections import Connections



def decode(x):
    if isinstance(x, str):
        return x
    return x.decode("utf-8")


class Project():
    """
    Class for using a project from RapidMiner Server that has been cloned locally. Use git for cloning the repository, then read and write calls can work on local resources. You need to use git commands to push changes that you make locally.
    """

    __NANOSECONDS_IN_A_DAY = 86400000000000
    _RM_HDF5_EXTENSION = ".rmhdf5table"
    _RM_RMP_EXTENSION = ".rmp"
    _METADATA_TYPES = {"NOMINAL": np.int8(1),
                       "INTEGER": np.int8(3),
                       "REAL": np.int8(4),
                       "TEXT": np.int8(5),
                       "BINOMINAL": np.int8(6),
                       "POLYNOMINAL": np.int8(7),
                       "FILE_PATH": np.int8(8),
                       "DATE_TIME": np.int8(9),
                       "DATE": np.int8(10),
                       "TIME": np.int8(11)}
    _LEGACY_TYPES = ["TEXT", "BINOMINAL", "FILE_PATH", "DATE"] # POLYNOMIAL handled separatly
    _METADATA_ROLES = ["BATCH", "CLUSTER", "ID", "LABEL", "OUTLIER", "PREDICTION", "WEIGHT"]

    __RM_MISSING_DATETIME_OR_TIME = 9223372036854775807 # constant value is coming from com.rapidminer.storage.hdf5.ExampleSetHdf5Writer.java#writeDateData/writeTimeData as of Long.MAX_VALUE/TimeColumn.MISSING_VALUE
    __PANDAS_MISSING_DATE = pd.to_numeric(pd.Series([datetime.utcfromtimestamp(0), None]), downcast="integer")[1]

    def __init__(self, path="."):
        """
        Initializes a reference to a locally cloned project. You need to clone a project from RapidMiner Server first (e.g. via git commands) to be able to use the methods of this instance.
        
        :param path: path to the local project repository root folder. It can be a relative path from the current working directory or an absolute path, . The default value points to the working directory.
        """
        if not os.path.exists(path):
            if path == "":
                msg_dir_part = "in the current directory"
            else:
                msg_dir_part = "at the specified path '%s'" % (os.path.dirname(path))
            raise ProjectException("Project '%s' does not exist %s. Please make sure you have a local copy of the project." % (os.path.basename(path), msg_dir_part))
        self.path = os.path.abspath(path)

    def read(self, path_or_buffer):
        """
        Reads a dataset from the local project repository into a pandas DataFrame. Note that only the new HDF5 format is supported, earlier RapidMiner Server data formats are not supported.
        
        :param path_or_buffer: this can either be a relative path inside the project (e.g. subfolder and file name), an absolute path, or a io.BytesIO stream. If a path is specified, the RapidMiner-specific HDF5 file extension can be omitted.
        """
        if isinstance(path_or_buffer, str):
            path_or_buffer = os.path.join(self.path, path_or_buffer)
            if not os.path.isfile(path_or_buffer):
                if (os.path.isfile(path_or_buffer + Project._RM_HDF5_EXTENSION)):
                    path_or_buffer = path_or_buffer + Project._RM_HDF5_EXTENSION
                else:
                    raise FileNotFoundError("File '%s' not found in the project." % (path_or_buffer))
        return Project.__read_data_safe(path_or_buffer)

    def write(self, df, path):
        """
        Writes a pandas DataFrame into the RapidMiner-specific HDF5 file format that the rest of the RapidMiner platform uses. Note that you need to explicitly commit and push your local changes to the remote project repository (e.g. via git commands) to make them available to the platform.
        
        :param path: relative path inside the project (e.g. subfolder and file name) specifying the target location or an absolute path. The RapidMiner-specific HDF5 file extension is automatically added to the filename, if it is missing.
        """
        path = os.path.join(self.path, path)
        if len(Path(path).suffix) == 0:
            path = path + Project._RM_HDF5_EXTENSION
        Project.__write_data_safe(df, path)

    def get_connections(self):
        """
        Returns the connection in that this project contains.
        """
        return Connections(self.path)
        
#####################
# Private functions #
#####################

    __from_ts_nanos = np.vectorize(lambda x: datetime.utcfromtimestamp(x/1e9) if x != Project.__RM_MISSING_DATETIME_OR_TIME else None)
    __from_ts_seconds_and_nanos = np.vectorize(lambda x,y: datetime.utcfromtimestamp(x+y/1e9) if x != Project.__RM_MISSING_DATETIME_OR_TIME else None)
    __hyp5_string_dtype = h5py.string_dtype()
    __h5py_reference_dtype = h5py.special_dtype(ref=h5py.Reference)
    
    def __get_numerical(x, typestr):
        if typestr == 'Integer':
            # NaN values are not allowed in int64
            if not np.isnan(np.min(x[:])):
                with x.astype('int64'):
                    return x[:]
            else:
                return x
        elif typestr in ('Date-Time', 'Date'):
            return Project.__from_ts_seconds_and_nanos(x[:],0)
        elif typestr == "Time":
            return Project.__from_ts_nanos(x[:])
        else:
            return x

    def __get_data(f, x):
        r = f[x]
        mapping = r.attrs.get('dictionary')
        if "additional" in r.attrs and r.attrs.get("type") == "Date-Time":
            additional = r.attrs.get('additional')
            additional = f[additional]
            return Project.__from_ts_seconds_and_nanos(r[:], additional[:])
        if mapping is None:
            return Project.__get_numerical(r, decode(r.attrs.get('type')))
        if isinstance(mapping,h5py.Reference):
            mapping = f[mapping]
        mapping = mapping[1:]
        g = lambda x:x-1
        return pd.Categorical.from_codes(g(r[()]), [decode(s) if not isinstance(s, str) else s for s in mapping], ordered=False)

    def __get_type(df_kind_char, hdf_attrs):
        """
        Decides type based on the pandas DataFrame type, the explicit HDF5 type and whether there is a positive nominal type that indicates binominal.
        """
        if "legacy_type" in hdf_attrs:
            for k in Project._METADATA_TYPES.keys():
                if Project._METADATA_TYPES[k] == hdf_attrs.get("legacy_type"):
                    return k.lower()
        hdf_type = decode(hdf_attrs.get("type"))
        if hdf_type in ("Date-Time", "Date"):
            meta_type = "date_time"
        elif hdf_type == "Time":
            meta_type = "time"
        elif df_kind_char in ('i', 'u'):
            meta_type = 'integer'
        elif df_kind_char in ('f'):
            meta_type = 'real'
        elif df_kind_char in ('M'):
            meta_type = 'date_time'
        elif df_kind_char in ('b') or "positive_index" in hdf_attrs:
            meta_type = 'binominal'
        else:
            meta_type = 'polynominal'
        return meta_type

    def __get_role(hdf_column):
        if "role" in hdf_column.attrs:
            r = decode(hdf_column.attrs.get("role"))
            if r in Project._METADATA_ROLES:
                return r.lower()
            elif r in ["SCORE", "METADATA"]:
                return decode(hdf_column.attrs.get("legacy_role"))
            else:
                raise ProjectException("Role '" + r +  "' not recognized. (Column '" + decode(hdf_column.attrs.get('name')) + "'.)" )
        return None
    
    def __read_data_from_input(f):
        numberOfColumns = f.attrs.get('columns')
        keys = ["a" + str(i) for i in range(0, numberOfColumns)]
        names = [decode(f[x].attrs.get('name')) for x in keys]
        data = [Project.__get_data(f, x) for x in keys]
        df = pd.DataFrame.from_dict(dict(zip(names, data)))
        metadata = dict(zip(names,
            [((Project.__get_type(df.dtypes[decode(f[x].attrs.get('name'))].kind,
                                 f[x].attrs)),
             Project.__get_role(f[x])) for x in keys]))
        set_metadata_without_warning(df, metadata)
        return df

    def __read_data_safe(filename_or_buffer):
        try:
            return Project.__read_data(filename_or_buffer)
        except OSError:
            raise ProjectException("Cannot read file. Not a valid rmhdf5table file format.")


    def __read_data(filename_or_buffer):
        with h5py.File(filename_or_buffer, 'r') as f:
            return Project.__read_data_from_input(f)

    def __get_desired_type(frame, column, name):
        desired_type = None
        if hasattr(frame, "rm_metadata") and name in frame.rm_metadata:
            desired_type = frame.rm_metadata[name][0].upper()
        else:
            if column.dtype.kind == 'O':
                desired_type = "NOMINAL"
            elif np.issubdtype(column.dtype, np.integer):
                desired_type = "INTEGER"
            elif column.dtype.kind == 'M':
                desired_type = "DATE_TIME"
            else:
                desired_type = "REAL"
        return desired_type

    def __create_dataset(f, column, desired_type, index):
        shortname = 'a'+str(index)
        if desired_type in ["NOMINAL", "BINOMINAL", "POLYNOMINAL", "TEXT", "FILE_PATH"]:
            cat = column.astype("category").cat
            dset = f.create_dataset(shortname, data = pd.to_numeric(cat.codes.apply(lambda x:x+1), downcast='integer'))
            mappingname = 'd'+str(index)
            mappingvals = cat.categories.values.astype(object)
            replacement = "NULL"
            while replacement in mappingvals:
                replacement = "\x00" + replacement
            mappingvals = np.concatenate([[replacement], mappingvals])
            if desired_type == "BINOMINAL" and len(mappingvals) > 3:
                raise TooManyBinomialValuesError("Column '%s' marked as binomial column in rm_metadata attribute, but has more there is more then two distinct values present.", )
            if len(mappingvals) <= 3:
                dset.attrs["dictionary"] = [str(v) for v in mappingvals]
                if desired_type == "BINOMINAL":
                    # positive index may change after read and write
                    dset.attrs['positive_index'] = np.int8(len(mappingvals) - 1)
            else:
                mset = f.create_dataset(mappingname, (len(mappingvals), ), dtype=Project.__hyp5_string_dtype)
                try:
                    mset[()] = mappingvals
                except TypeError:
                    # fixes error: TypeError: Can't implicitly convert non-string objects to strings (caused by some Python environments)
                    mset[()] = [str(v) for v in mappingvals]
                dset.attrs.create("dictionary", mset.ref, dtype=Project.__h5py_reference_dtype)
            dset.attrs['type'] = "Nominal"
        elif desired_type == "INTEGER":
            dset = f.create_dataset(shortname, data = column.astype("int64"))
            dset.attrs['type'] = "Integer"
        elif desired_type in ["DATE", "TIME", "DATE_TIME"]:
            if column.dtype.kind == "M":
                nanoseconds = pd.to_numeric(column.dt.tz_localize(None), downcast="integer")
            else:
                nanoseconds = pd.to_numeric(column, downcast="integer")
            if desired_type == "TIME":
                missing = nanoseconds==Project.__PANDAS_MISSING_DATE
                nanoseconds.loc[missing] = Project.__RM_MISSING_DATETIME_OR_TIME
                nanoseconds.loc[~missing] = nanoseconds.loc[~missing] % int(Project.__NANOSECONDS_IN_A_DAY)
                dset = f.create_dataset(shortname, data = nanoseconds)
                dset.attrs['type'] = "Time"
            else:
                seconds = nanoseconds.apply(lambda x: x // int(1e9) if x != Project.__PANDAS_MISSING_DATE else Project.__RM_MISSING_DATETIME_OR_TIME)
                dset = f.create_dataset(shortname, data = seconds )
                additionalname = shortname+"a"
                aset = f.create_dataset(additionalname, data = (nanoseconds % int(1e9)).astype("int32"))
                dset.attrs.create("additional", aset.ref, dtype=Project.__h5py_reference_dtype)
                dset.attrs['type'] = "Date-Time"
        else:
            dset = f.create_dataset(shortname, data = column.astype("float64"))
            dset.attrs['type'] = "Real"
        return dset

    def __to_column_role(role):
        if role.startswith("confidence"):
            return ("SCORE", role)
        elif role.upper() in Project._METADATA_ROLES:
            return (role.upper(), None)
        else:
            return ("METADATA", role)

    def __set_common_column_attributes(frame, dset, desired_type, name):
        dset.attrs['name'] = name
        if desired_type in Project._LEGACY_TYPES:
            dset.attrs['legacy_type'] = Project._METADATA_TYPES[desired_type]
        if hasattr(frame, "rm_metadata") and name in frame.rm_metadata:
            md = frame.rm_metadata[name]
            if len(md) > 1 and md[1]:
                role, legacy_role = Project.__to_column_role(md[1])
                dset.attrs['role'] = role
                if legacy_role is not None:
                    dset.attrs['legacy_role'] = legacy_role

    def __write_column(f, frame, column, name, index):
        desired_type = Project.__get_desired_type(frame, column, name)
        try:
            dset = Project.__create_dataset(f, column, desired_type, index)
        except ValueError as e:
            raise ValueConversionError("Cannot write output file with the desired format. Please review rm_metadata of the pandas DataFrame. Cause: " + str(e))
        Project.__set_common_column_attributes(frame, dset, desired_type, name)

    def __write_data_safe(data, filename):
        if not isinstance(data, pd.DataFrame):
            raise TypeError("'data' attribute of write_data is not pandas DataFrame.")
        if not os.path.exists(os.path.dirname(filename)):
            raise ProjectException("Cannot write file. Parent directory '%s' does not exists." % (os.path.dirname(filename)))
        if hasattr(data, "rm_metadata"):
            md = data.rm_metadata
            for key in md.keys():
                if md[key][0] is not None and md[key][0].upper() not in Project._METADATA_TYPES.keys():
                    raise ProjectException("%s is not a valid type in rm_metadata." % (md[key][0]))
        Project.__write_data(data, filename)

    def __write_data(data, filename): #todo: error handling
        frame = data
        # without this call you get fixed length strings which can only be ascii
        with h5py.File(filename, 'w') as f:
            shape = frame.shape
            f.attrs['rows'] = np.int32(shape[0])
            f.attrs['columns'] = np.int32(shape[1])
            for i in range(0, len(frame.columns)):
                name = str(frame.columns[i])
                column = frame.iloc[:,i]
                Project.__write_column(f, frame, column, name, i)



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import requests
import base64
import tempfile
import xml.etree.ElementTree as et
import pandas as pd
import getpass
try:
    import cPickle as pickle
except:
    import pickle
from time import sleep
from .connections import Connections
from .connector import Connector
from .utilities import ServerException
from .utilities import GeneralException
from .utilities import extract_json
from .utilities import Version
from .utilities import VersionException
from .utilities import _is_docker_based_deployment
from .resources import RepositoryLocation, ProjectLocation
from .serdeutils import read_example_set, write_example_set, is_file_object
from .project import Project
from functools import partial
import warnings
import io
import logging
import textwrap
import zeep
import os
import sys
import json
import re
from pathlib import Path

from .. import __version__

def get_server():
    if _is_docker_based_deployment():
        default_url = "http://rm-server-svc:8080"
        if "SERVER_BASE_URL" in os.environ:
            default_url = os.environ["SERVER_BASE_URL"]
        return Server(url=default_url)
    else:
        return Server()

class Server(Connector):
    """
    Class for using a local or remote RapidMiner Server instance directly. You can read from and write to the Server repository and you can execute processes using the scalable Job Agent architecture.
    """
    __POLL_INTERVAL_SECONDS = 6
    __PROCESS_PATH_TRIES = 3
    __WEBSERVICE_PROCESS_XML = \
        """<?xml version="1.0" encoding="UTF-8"?><process version="9.3.000">
          <context>
            <input/>
            <output/>
            <macros/>
          </context>
          <operator activated="true" class="process" compatibility="9.3.000" expanded="true" name="Process">
            <process expanded="true">
              <operator activated="true" class="python_scripting:repository_service" compatibility="9.3.000" expanded="true" height="68" name="Repository Access" width="90" x="179" y="34"/>
              <connect from_port="input 1" to_op="Repository Access" to_port="file"/>
              <connect from_op="Repository Access" from_port="output" to_port="result 1"/>
              <portSpacing port="source_input 1" spacing="0"/>
              <portSpacing port="source_input 2" spacing="0"/>
              <portSpacing port="sink_result 1" spacing="0"/>
              <portSpacing port="sink_result 2" spacing="0"/>
            </process>
          </operator>
        </process>
        """
    __WEBSERVICE_DESCRIPTOR_XML = \
        """<?xml version="1.0" encoding="UTF-8"?>
        <exported-process>
            <mime-type>application/json</mime-type>
            <output-format>JSON</output-format>
            <process-entry>PROCESS_ENTRY_PATH</process-entry>
            <xslt-entry/>
            <parameter-mappings/>
            <properties>
                <name>service:Repository Service</name>
            </properties>
            <data-source-input>
                <name>service:Repository Service</name>
            </data-source-input>
        </exported-process>
        """
    __WELCOME_INFO = \
"""Web service backend for the Server Python API on this Server is not installed, this
is probably because you are using this API for the first time. This
backend will be installed once and can be used forever afterwards.
Please use either RapidMiner Studio or the RapidMiner Server web interface
to create a dedicated folder for this API backend process that is shared 
among all the users.
Please read the documentation (README) on how to create this folder. If
you are the sole user of this Server, you may skip the creation of this
folder and simply use your home folder. In that case note that if you
want to allow other users later to use the API, you will need to give
them access to the process created by this operation."""
    __WELCOME_LINE_LENGTH = 70
    __TAB = "    "
    __SHARED_PROCESS_FOLDER = "/shared"

    def __init__(self, url='http://localhost:8080', username=None, password=None, size_limit_kb=50000,
                 webservice="Repository Service", processpath=None, tempfolder=None, install=True, verifySSL=True,
                 logger=None, loglevel=logging.INFO):
        """
        Initializes a new connector to a local or remote Rapidminer Server instance. It also installs the auxiliary web service required by this library to be able to interact with the Server repository directly.

        Arguments:
        :param url: Server url path (hostname and port as well)
        :param username: user to use Server with
        :param password: password for the username. If not provided, you will need to enter it.
        :param size_limit_kb: this setting is only applied when using a (legacy) repository and not a project! Maximum number of kilobytes that are allowed to be read from or writing to Server. Reading or writing large objects may degrade Server's performance or lead to out of memory errors. Default value is 50000.
        :param webservice: this API requires an auxiliary process installed as a web service on the Server instance. This parameter specifies the name for this web service. The web service is automatically installed if it has not been.
        :param processpath: path in the repository where the process behind the web service will be saved. If not specified, a user prompt asks for the path, but proposes a default value. Note that you may want to make this process executable for all users.
        :param tempfolder: repository folder on Server that can be used for storing temporary objects by run_process method. Default value is "tmp" inside the user home folder. Note that in case of certain failures, you may need to delete remaining temporary objects from this folder manually.
        :param install: boolean. If set to false, web service installation step is completely skipped. Default value is True.
        :param verifySSL: either a boolean, in which case it controls whether we verify the server's TLS certificate, or a string, in which case it must be a path to a CA bundle to use. Default value is True.
        :param logger: a Logger object to use. By default a very simple logger is used, with INFO level, logging to stdout.
        :param loglevel: the loglevel, as an int value. Common values are defined in the standard logging module. Only used, if logger is not defined.
        """
        super(Server, self).__init__(logger, loglevel)
        # URL of the Rapidminer Server
        if url.endswith("/"):
            url = url[:-1]
        self.server_url = url
        if not _is_docker_based_deployment():
            # RapidMiner Server Username
            if username is None:
                username = input('Username: ')
            self.username = username
            # RapidMiner Server Password
            if password is None:
                password = getpass.getpass(prompt='Password: ')
            self.__password = password
        else:
            self.username = None
            self.__password = None
        self.size_limit_kb = size_limit_kb
        self.webservice = webservice
        self.__processpath = processpath
        if tempfolder is None:
            self.__tempfolder = "/home/" + self.__username() + "/tmp/"
        else:
            self.__tempfolder = tempfolder
            self.__tempfolder += "/" if not self.__tempfolder.endswith("/") else ""
        self.__install = install
        self.__verifySSL = verifySSL
        self.__user_agent =  "RapidMiner Python Package " + str(__version__)

        # Connect to the RM Server
        self.__connect()

        # Test and install required web service if it does not exist
        if self.__install:
            self.__test_and_install()

####################
# Public functions #
####################

    def read_resource(self, path, project=None):
        """
        Reads one or more resources from the specified Server repository locations. Does not allow the retrieval of objects larger than the limit settings allow (size_limit_kb, limit in kilobytes).

        Arguments:
        :param path: the path(s) to the resource(s) inside Server repository. Multiple paths can be specified as list or tuple. A path can be a string, a rapidminer.RepositoryLocation or a rapidminer.ProjectLocation object.
        :param project: optional project name. If this argument is defined, the specified paths are resolved in this project. If rapidminer.ProjectLocation objects are specified in the first argument, their project settings override the project set by this argument.
        :return: the requested resource(s). Datasets are returned as pandas DataFrames. Python native objects are returned as Python objects, other object types may be returned as bytes or BytesIO. If multiple inputs are specified, the same number of inputs will be returned in a tuple. Otherwise, the return value is a single object.
         """
        if not ((isinstance(path, tuple) or isinstance(path, list))):
            path = [path]
            single_input = True
        else:
            if len(path) == 0:
                return None
            single_input = False
        resources = []
        for inp in path:
            this_project = project
            if isinstance(inp, RepositoryLocation):
                inp = inp.to_string(with_prefix=False)
            elif isinstance(inp, ProjectLocation):
                this_project = inp.project
                inp = inp.path
            elif not isinstance(inp, str):
                raise ServerException("Input path should be 'str' or 'rapidminer.RepositoryLocation or 'rapidminer.ProjectLocation object, not '" + str(type(inp)) + "'.")
            if this_project:
                resources.append(self.__read_project(this_project, inp))
            else:
                resources.append(self.__read_repository(inp))
        if single_input:
            return resources[0]
        else:
            return tuple(resources)

    def write_resource(self, resource, path):
        """
        Writes pandas DataFrame(s) or other objects to the Server repository.

        Arguments:
        :param resource: a resource can be a pandas DataFrame, a pickle-able python object or a file-like object. Multiple DataFrames or other objects can be specified as list or tuple. A path can be a string or a rapidminer.RepositoryLocation object.
        :param path: the target path(s) to the resource(s) inside Server repository. The same number of path values are required as the number of resources.
        """
        if not ((isinstance(resource, tuple) or isinstance(resource, list))):
            resource = [resource]
        if not ((isinstance(path, tuple) or isinstance(path, list))):
            path = [path]
        if len(resource) != len(path):
            raise ValueError("resource and path must contain the same number of values")
        for obj in resource:
            if sys.getsizeof(obj) > self.size_limit_kb * 1024:
                raise ServerException(
                    "Specified object is larger than the current size limit for objects written to Server (" 
                    + str(self.size_limit_kb) + " KB). Increase size_limit_kb setting of this class if you "
                    + "want write larger objects. Note that writing very large objects via this API may degrade Server performance or cause it to run out of memory.")
        for obj, out, idx in zip(resource, path, range(len(resource))):
            if isinstance(out, RepositoryLocation):
                out = out.to_string(with_prefix=False)
            elif not isinstance(out, str):
                raise ServerException("Output path should be 'str' or 'rapidminer.RepositoryLocation object, not '" + str(type(out)) + "'.")
            post_url = self.server_url + "/api/rest/process/" + self.webservice + "?"
            if type(obj) == pd.DataFrame:
                csv_stream = io.StringIO()
                md_stream = io.StringIO()
                write_example_set(self._copy_dataframe(obj), csv_stream, md_stream)
                data = [{"extension": "csv-encoded", "content": csv_stream.getvalue()},
                        {"extension": "pmd-encoded", "content": md_stream.getvalue()}]
            elif is_file_object(obj):
                raise ServerException("Writing file objects is not supported. Read the file content into the memory before calling this method if you want to write it to the repository.")
            else:
                bin_stream = io.BytesIO()
                pickle.dump(obj, bin_stream)
                data = [{"extension": "bin", "content": base64.b64encode(bin_stream.getvalue()).decode("utf-8")}]
            r = self.__send_request(partial(requests.post, post_url, json={"command": "write_resource", "library_version": __version__, "path": out, "data": data}),
                                    lambda s: "Failed to save input no. " + str(idx+1) + ", status: " + str(s)
                                    + (". The web service backend may have been deleted, please try to use a new Server class." if s == 404 else ""))
            response = extract_json(r)
            self.__check_extension_version(response)

    def run_process(self, path, inputs=[], macros={}, queue="DEFAULT", ignore_cleanup_errors=True, project=None):
        """
        Runs a RapidMiner process and returns the result(s).

        Arguments:
        :param path: path to the RapidMiner process in the Server repository. It can be a string or a rapidminer.RepositoryLocation object.
        :param inputs: inputs used by the RapidMiner process, an input can be a pandas DataFrame, a pickle-able python object or a file-like object.
        :param macros: optional dict that sets the macros in the process context according to the key-value pairs, e.g. macros={"macro1": "value1", "macro2": "value2"}
        :param queue: the name of the queue to submit the process to. Default is DEFAULT.
        :param ignore_cleanup_errors: boolean. Determines if any error during temporary data cleanup should be ignored or not. Default value is True.
        :param project: optional project name. If this argument is defined, the specified paths are resolved in this project. If rapidminer.ProjectLocation objects are specified in the first argument, their project settings override the project set by this argument. Note that when using projects, the inputs parameter is ignored, as direct write is not supported for versioned projects. Also, the method does not return outputs in this case.
        :return: the results of the RapidMiner process. It can be None, or a single object, or a tuple. One result may be a pandas DataFrame, a pickle-able Python object or a file-like object. When a project is used, no output is returned, as that would need direct write to a versioned project that is not supported.
        """
        this_project = project
        if isinstance(path, RepositoryLocation):
            path = path.to_string(with_prefix=False)
        elif isinstance(path, ProjectLocation):
            this_project = path.project
            path = path.path
        elif not isinstance(path, str):
            raise ServerException("Process path should be 'str' or 'rapidminer.RepositoryLocation object, not '" + str(type(path)) + "'.")
        if inputs is not None and not ((isinstance(inputs, tuple) or isinstance(inputs, list))):
            inputs = [inputs]
        if this_project:
            process_xml = self.__read_process_from_project(this_project, path)
            # location will not be submitted
            path = ProjectLocation(this_project, path).to_string()
        else:
            process_xml = self.__read_process_xml(path)
        root = et.fromstring(process_xml)
        temp_resources = []
        context = {}
        try:
            if this_project:
                if inputs:
                    print("When running a process from a project, inputs cannot be specified.")
            else:
                if inputs:
                    input_resources = [self.__tempfolder + next(tempfile._get_candidate_names()) for _ in inputs]
                    temp_resources += input_resources
                    self.write_resource(inputs, input_resources)
                    # add input locations in process xml
                    context["inputLocations"] = input_resources
                # find connected output ports, add locations to process xml
                output_resources = []
                for wire in root.find('operator').find('process').findall('connect'):
                    if wire.attrib['to_port'].startswith('result '):
                        output_resources.append(self.__tempfolder + next(tempfile._get_candidate_names()))
                if output_resources:
                    context["outputLocations"] = output_resources
                temp_resources += output_resources
            # set macros in process xml
            if macros:
                macros_dict = {}
                for key, value in macros.items():
                    macros_dict[key] = value
                context["macros"] = macros_dict
            r = self.__submit_process_xml(queue, process_xml, path, context)
            jobid = r.json()["id"]
            self.log("Submitted process with job id: " + str(jobid))
            self.__wait_for_job(jobid)
            if this_project:
                return None
            else:
                if len(output_resources) == 1:
                    output_resources = output_resources[0]
                return self.read_resource(output_resources)
        finally:
            if ignore_cleanup_errors:
                try:
                    self.__delete_resource(temp_resources)
                except Exception as e:
                    strfile = "entry" if len(temp_resources) == 1 else "entries"
                    message = e.message if hasattr(e, 'message') else str(e)
                    self.log("Could not delete the following temporary " + strfile + ", error: " + message, level=logging.WARNING)
                    self.log("\n".join(t for t in temp_resources), level=logging.WARNING)
            else:
                self.__delete_resource(temp_resources)

    def get_queues(self):
        """
        Gets information of the available queues in the Server instance.

        :return: a JSON array of objects representing each queue with its properties
        """
        get_url = self.server_url + "/executions/queues?"
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get queues, status: " + str(s))
        return r.json()

    def get_projects(self):
        """
        Gets information of the available projects in the AI Hub instance.

        :return: a JSON array of objects representing each repository with its properties
        """
        get_url = self.server_url + "/executions/repositories?"
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get projects, status: " + str(s))
        return r.json()
    
    def get_connections(self):
        """
        Read the connections from the AI Hub repository.
        
        :return: Connections object listing connections from the AI Hub repository. Note that values of encrypted fields are not available (values will be None). Use AI Hub Vault to securely store and retrieve these values instead  
        """
        return Connections(path=None, server=self)
        
        
#####################
# Private functions #
#####################

    def _get_vault_info(self, location):
        """
        Load all server vault entries for a repository location.
        
        :param location: the location to read the vault information for. Note that there is a different syntax depending on the target repository type. For Projects, use git://reponame.git/Connections/My Connection.conninfo. For AI Hub repository, use /Connections/My Connection
        """
        if isinstance(location, ProjectLocation):
            location = location.to_string(with_prefix=True)
        elif not isinstance(location, str):
            raise ServerException("Location must be a 'str' or 'rapidminer.ProjectLocation object, not '" + str(type(inp)) + "'.")
        get_url = self.server_url + "/executions/connections/vault?location=" + location
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get vault info, status: " + str(s))
        return r.json()

    def _get_project_info(self, project_name):
        """
        Read the information for a project from AI Hub.
        
        :param project_name: specifies the project
        :return: info in JSON format for the project
        """
        get_url = self.server_url + "/executions/repositories/" + project_name
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get project info" 
                                + (": No project exists with the name '" + project_name + "', provide a valid project name" 
                                   if s == 404 else ", status: " + str(s)))
        return r.json()

    def _get_connections_info(self):
        """
        Read the connections from the AI Hub repository.
        
        :return: connections in JSON format
        """
        get_url = self.server_url + "/api/rest/repository/connections"
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get connections list, status: " + str(s))
        return r.json()

    def _read_connection_info(self, location):
        get_url = self.server_url + "/executions/connections/detail?location=" + location
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get connection details, status: " + str(s))
        return r.json()

    def __username(self):
        if _is_docker_based_deployment():
            if "JUPYTERHUB_USER" in os.environ:
                return os.environ["JUPYTERHUB_USER"]
            else:
                raise ServerException("Internal exception: JUPYTERHUB_USER environment variable not set.")
        else:
            return self.username

    def __connect_idp(self):
        api_token = os.getenv('JUPYTERHUB_API_TOKEN')
        api_url = os.getenv('JUPYTERHUB_API_URL')
        api_url = re.sub(r"(https?:/)",r"\g<1>/", re.sub("//*", "/", api_url + '/whoami'))

        self.log("Getting up-to-date access token from: " + api_url, level=logging.DEBUG)
        r = requests.get(api_url, headers={'Authorization': 'token %s' % api_token})
        r.raise_for_status()
        jwt = r.json()['auth_state']['rms_jwt_idToken']
        # Bearer Authorization header
        return { 'Authorization' : 'Bearer %s' %  jwt }

    def __connect_basic_auth_header(self):
        # Encode the basic Authorization header
        userAndPass = base64.b64encode(bytes(self.username + ":" + self.__password, "utf-8")).decode("ascii")
        header = { 'Authorization' : 'Basic %s' %  userAndPass }
        # dummy call to check credentials
        r = self.__send_request(partial(requests.get, url=self.server_url + '/api/rest/instance'),
                                error_fn=lambda s:
                                "Connection error, status: " + str(s) + "."
                                + (" Make sure that you entered a valid username and password." if s == 401 else ""),
                                reconnect=False, headers_fn=lambda: header)
        if "Access denied" in r.text:
            raise ServerException("Connection error: Access denied")
        return header

    def __connect(self):
        if _is_docker_based_deployment():
            self.auth_header = self.__connect_idp()
        else:
            # currently, only user/password is supported
            self.auth_header = self.__connect_basic_auth_header()
        self.log("Successfully connected to the Server")

    def __print_welcome_msg(self):
        # text displayed to explain why the user needs to choose a path where the web service process is installed to
        print(self.__TAB + ("*" * self.__WELCOME_LINE_LENGTH))
        print(self.__TAB + ("\n" + self.__TAB).join(textwrap.wrap(self.__WELCOME_INFO, self.__WELCOME_LINE_LENGTH)))
        print(self.__TAB + ("*" * self.__WELCOME_LINE_LENGTH))
        
    def __test_and_install(self):
        # test if web service exists
        post_url = self.server_url + "/api/rest/process/" + self.webservice + "?"
        shared_folder_exists = self.__is_folder(self.__SHARED_PROCESS_FOLDER)
        r = self.__send_request(partial(requests.post, post_url, json={"command": "test", "library_version": __version__}))
        if r.status_code == 404:
            if not shared_folder_exists:
                self.__print_welcome_msg()
            self.log("Web service is not installed, installing it with the name '" + self.webservice + "'...")
            
            # first try to install into /shared folder (available since Server 9.6)
            installed = False
            if shared_folder_exists:
                processpath_to_test = self.__SHARED_PROCESS_FOLDER + "/repository_api/" + self.webservice
                # if /shared folder exists, install should succeed, nevertheless, we give the user a chance to
                # provide a custom folder to install to if they get an unauthorized error
                installed = self.__install_webservice(processpath_to_test)
                if not installed:
                    self.__print_welcome_msg()
            if not installed:
                default_webservice_path = "/repository_api/" + self.webservice
                # if there is no process path specified, try to get it from user input a couple of times
                tries = self.__PROCESS_PATH_TRIES if self.__processpath is None else 1
                for _ in range(tries):
                    processpath_to_test = input("Please enter full repository path for installing the process - it will be accessible by all users"
                                            + " [" + default_webservice_path + "]: ") if self.__processpath is None else self.__processpath
                    if processpath_to_test.strip() == "":
                        processpath_to_test = default_webservice_path
                    if self.__is_folder(processpath_to_test):
                        processpath_to_test += ("" if processpath_to_test.endswith("/") else "/") + self.webservice
                        self.log("You entered a folder name. Trying to save the process in that folder: " + processpath_to_test, logging.WARNING)
                    if self.__install_webservice(processpath_to_test):
                        break
            self.__processpath = processpath_to_test
            # Re-test installed service
            r = self.__send_request(partial(requests.post, post_url, json={"command": "test", "library_version": __version__}),
                                    lambda s: "Test of installed web service failed, status: " + str(s))
            response = extract_json(r)
            self.__check_extension_version(response)
            self.log("Web service backend installed successfully")
        elif r.status_code == 200:
            response = extract_json(r)
            self.__check_extension_version(response)
            self.log("Web service backend exists")
        else:
            raise ServerException("Web service test failed with unexpected error, status: " + r.status_code \
                                  + ". Make sure that the web service with the name '" + self.webservice + ' is installed.')

    def __read_process_xml(self, path):
        get_url = self.server_url + "/api/rest/resources" + path
        r = self.__send_request(partial(requests.get, get_url),
                                lambda s: "Failed to get process \"" + path + "\", status: " + str(s))
        return r.text

    def __submit_process_xml(self, queue, process, location, context):
        post_url = self.server_url + "/executions/jobs?"
        body = {
            "queueName": queue,
            "process": base64.b64encode(bytes(process, "utf-8")).decode("ascii"),
            "location": location,
            "context": context
        }
        return self.__send_request(partial(requests.post, url=post_url, json=body),
                                   lambda s: "Failed to submit process, status: " + str(s))

    __JOB_STATE_ERROR = ("TIMED_OUT", "STOPPED", "ERROR")
    __JOB_STATE_SUCCESS = ("FINISHED")

    def __wait_for_job(self, jobid):
        while True:
            sleep(self.__POLL_INTERVAL_SECONDS)
            get_url = self.server_url + "/executions/jobs/" + jobid
            r = self.__send_request(partial(requests.get, get_url),
                                    lambda s: "Error during getting job status, job id: " + jobid + ", status: " + str(s))
            r = r.json()
            if r["state"] in self.__JOB_STATE_ERROR:
                raise ServerException("Job finished with error state: " + r["state"] + ", " + Server.__format_job_error(r))
            elif r["state"] in self.__JOB_STATE_SUCCESS:
                return

    def __format_job_error(response):
        # TODO: improve
        return "Unknown error" if "error" not in response else response["error"]["type"] + ": " + response["error"]["title"] + ": " + response["error"]["message"]

    def __delete_resource(self, resource_paths):
        post_url = self.server_url + "/api/rest/process/" + self.webservice + "?"
        for path in resource_paths:
            self.__send_request(partial(requests.post, post_url, json={"command": "delete_resource", "library_version": __version__, "path": path}),
                                lambda s: "Failed to delete path \"" + path + "\", status: " + str(s))

    def __install_webservice(self, path):
        unauthorized_to_save_msg = "Unauthorized to save. "
        try:
            self.__post_process(path, self.__WEBSERVICE_PROCESS_XML, unauthorized_to_save_msg)
            self.__make_public(path)
            self.__post_service(self.webservice, self.__WEBSERVICE_DESCRIPTOR_XML.replace("PROCESS_ENTRY_PATH", path),
                               "Your user could not create the web service backend. Please ask an administrator to"
                                       + " create a web service from the following process: " + path + ". ")
            return True
        except ServerException as e:
            if str(e).startswith(unauthorized_to_save_msg):
                self.log("Could not save process to location: " + path + ". Please use a path where your user can write to.", logging.ERROR)
                self.log("Error: " + str(e)[len(unauthorized_to_save_msg):], logging.ERROR)
                return False
            else:
                raise e

    def __post_process(self, path, process, unauthorized_error_msg):
        post_url = self.server_url + "/api/rest/resources" + path
        return self.__send_request(partial(requests.post, post_url, data=process),
                                   error_fn=lambda s:
                                       (unauthorized_error_msg if s == 403 else "")
                                       + "Failed to save process to repository path '" + path + "', status: " + str(s),
                                   accepted_status_codes=[201], headers_fn=self.__get_headers_with_content_type)

    def __make_public(self, path):
        client = self.__get_soap_client()
        try:
            res = client.service.setAccessRights(path, {
                'execute': 'GRANT',
                'group': 'users',
                'read': 'GRANT',
                'write': 'IGNORE'
            })
        finally:
            self.__destroy_soap_client(client)
        if hasattr(res, "status") and res["status"] != 0:
            self.log("Could not give permission on the backend process to 'users' group, status : " + str(res["status"])
                + ((", error message: " + res["errorMessage"]) if hasattr(res, "errorMessage") else "")
                + ". Please set it manually, or ask an administrator to do it.", logging.WARNING)

    def __post_service(self, service_name, descriptor, unauthorized_error_msg):
        post_url = self.server_url + "/api/rest/service/" + service_name
        return self.__send_request(partial(requests.post, post_url, data=descriptor),
                                   error_fn=lambda s:
                                       "Failed to install web service with the name '" + service_name + "', status: " + str(s) + "." +
                                       (" Reason: " if s in [403,500] else "") +
                                       (unauthorized_error_msg if s == 403 else "") +
                                       ("Internal server error. You may not have sufficient privileges to complete this action. Ask an administrator to install the web service." if s == 500 else ""))

    def __is_folder(self, path):
        client = self.__get_soap_client()
        try:
            res = client.service.getFolderContents(path)
            if hasattr(res, "status") and res["status"] == 0:
                return True
            else:
                return False
        finally:
            self.__destroy_soap_client(client)

    def __get_soap_client(self):
        session = requests.Session()
        for k, v in self.auth_header.items():
            session.headers[k] = v
        session.headers["User-Agent"] = self.__user_agent
        client = zeep.Client(self.server_url + "/api/soap/RepositoryService?wsdl", transport=zeep.transports.Transport(session=session))
        return client

    def __destroy_soap_client(self, client):
        client.transport.session.close()

    def __get_headers(self):
        head = self.auth_header.copy()
        head["User-Agent"] = self.__user_agent
        return head

    def __get_headers_with_content_type(self):
        head = self.__get_headers().copy()
        head['Content-Type'] = 'application/vnd.rapidminer.rmp+xml'
        return head

    def __send_request(self, request, error_fn=None, accepted_status_codes=[200], reconnect=True, headers_fn=None):
        if headers_fn is None:
            headers_fn = self.__get_headers
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Unverified HTTPS request")
                response = request(headers=headers_fn(), verify=self.__verifySSL)
                if reconnect and response.status_code == 401:
                    # token may have expired, try to reconnect:
                    self.log("Session may have expired. Trying to reconnect to the server...")
                    self.__connect()
                    response = request(headers=headers_fn(), verify=self.__verifySSL)
        except requests.exceptions.SSLError as e:
            if "SSL: CERTIFICATE_VERIFY_FAILED" in str(e):
                raise ServerException("SSL verification error while connecting to the server." +
                                      " Probably you did not configured properly the certificate of your RapidMiner Server." +
                                      " By specifying the 'verifySSL=False' or 'verifySSL=/path/to/your/certificate' parameter for the rapidminer.Server object," +
                                      " you can dismiss this error, though this is not recommended." +
                                      " For more details consult the documentation.") from e
            else:
                raise e
        if response.status_code not in accepted_status_codes and error_fn is not None:
            raise ServerException(error_fn(response.status_code))
        return response

    def __check_extension_version(self, response, typeColumn="type", valueColumn="value"):
        extensionVersion = None
        try:
            if isinstance(response, dict):
                if response[typeColumn] == "version_info":
                    extensionVersion = response[valueColumn]
            elif isinstance(response, list):
                for row in response:
                    if row[typeColumn] == "version_info":
                        extensionVersion = row[valueColumn]
                        break
        except (TypeError, KeyError):
            # the response is in the wrong (legacy) format
            pass
        if extensionVersion is None or not Version(extensionVersion).is_at_least(Version("9.7.0")):
            raise VersionException("RapidMiner Server", "9.7.0 or newer")

    def __read_project(self, project, path):
        get_url = self.server_url + "/executions/repositories/" + project + "/resources/master/" + path
        try:
            r = self.__send_request(partial(requests.get, get_url),
                lambda s: "Failed to get resource " + project + "/" + path + ", status: " + str(s))
        except ServerException as e:
            # re-try with our hdf5 file extension if it has not been specified
            if str(e).endswith("404") and len(Path(path).suffix) == 0:
                r = self.__send_request(partial(requests.get, get_url + Project._RM_HDF5_EXTENSION),
                    lambda s: "Failed to get resource " + project + "/" + path + ", status: " + str(s))
            else:
                raise e
        if path.endswith(Project._RM_HDF5_EXTENSION):
            return Project().read(io.BytesIO(r.content))
        elif len(Path(path).suffix) == 0:
            try:
                return Project().read(io.BytesIO(r.content))
            except OSError as e:
                # most likely not a hdf5 file
                pass
        return r.content

    def __read_repository(self, path):
        post_url = self.server_url + "/api/rest/process/" + self.webservice + "?"
        r = self.__send_request(partial(requests.post, post_url, json={"command": "read_resource", "library_version": __version__, "path": path, "size_limit_kb": self.size_limit_kb}),
                                lambda s: "Failed to read input \"" + path + "\", status: " + str(s) 
                                + (". The web service backend may have been deleted, please try to use a new Server class." if s == 404 else ""))
        response = extract_json(r)
        self.__check_extension_version(response, typeColumn="extension", valueColumn="content")
        if not isinstance(response, list) or len(response) not in [1,2,3]:
            raise ServerException("Invalid response from server: " + response)
        csv_data = None
        metadata = None
        for row in response:
            try:
                ext = row["extension"]
                content = row["content"]
                if ext == "csv-encoded":
                    csv_data = io.StringIO(content)
                elif ext == "pmd-encoded":
                    metadata = io.StringIO(content)
                elif ext == "bin":
                    try:
                        obj = pickle.load(io.BytesIO(base64.b64decode(content)))
                    except Exception as exc:
                        raise GeneralException("Error while trying to load pickled object (note that Python 2 objects may not be readable): " + str(exc))
                    return obj
                elif ext == 'fo':
                    return io.BytesIO(base64.b64decode(content.encode("utf-8"))) # reads the file to memory
            except KeyError as e:
                raise ServerException("The response from the server differs from the expected.") from e
        if csv_data is not None:
            if metadata is None:
                raise ServerException("No metadata found.")
            dataframe = read_example_set(csv_data, metadata)
            return dataframe
        if metadata is not None and csv_data is None:
            raise ServerException("No data found.")

    def __read_process_from_project(self, project, path):
        get_url = self.server_url + "/executions/repositories/" + project + "/resources/master/" + path
        try:
            r = self.__send_request(partial(requests.get, get_url),
                lambda s: "Failed to find process at " + project + "/" + path + ", status: " + str(s))
        except ServerException as e:
            # re-try with rmp file extension if it has not been specified
            if str(e).endswith("404") and len(Path(path).suffix) == 0:
                r = self.__send_request(partial(requests.get, get_url + Project._RM_RMP_EXTENSION),
                                        lambda s: "Failed to find process at " + project + "/" + path + ", status: " + str(s))
            else:
                raise e
        return r.text



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import shutil
import os
import subprocess
import tempfile
import glob
import sys
import logging
from threading import Thread
import threading
import platform
import io
import pandas
try:
    import cPickle as pickle
except:
    import pickle
from .utilities import __STDOUT_ENCODING__
from .connector import Connector
from .resources import Resource
from .resources import File
from .resources import RepositoryLocation
from .utilities import GeneralException
from .utilities import __open__
from .utilities import Version
from .utilities import VersionException
from .serdeutils import read_example_set, write_example_set

from .. import __version__

class StudioException(Exception):
    def __init__(self, msg=""):
        super(Exception, self).__init__(msg)

class Studio(Connector):
    """
    Class for using a locally installed RapidMiner Studio instance. You can read from and write to the repositories defined in Studio (and use even remote repositories this way) and you can execute processes.
    """
    __TMP_OUTPUT_DIR_PREFIX= "rapidminer-scripting-output-"
    __TMP_INPUT_DIR_PREFIX="rapidminer-scripting-inputs-"
    __VERSION_MSG="RAPIDMINER_VERSION="
    ___EXIT_CODE_MSG="EXIT_CODE="
    __RAPIDMINER_ERROR_MSG="RAPIDMINER_ERROR_MSG="
    __RAPIDMINER_ERROR_MSG_FIRST_LINE="RAPIDMINER_ERROR_MSG_FIRST_LINE="
    __SCRIPTS_SUBDIR="scripts"

    def __init__(self, studio_home=None, logger=None, loglevel=logging.INFO, rm_stdout=None, override_python_binary=False):
        """
        Initializes a new connector to a local Rapidminer Studio instance. Every command will launch a new Studio instance, executing the required operations in batch mode.

        Arguments:
        :param studio_home: path to installation directory of the Rapidminer Studio. If None, the location will be taken from the RAPIDMINER_HOME environment variable if defined, or the current directory, if RAPIDMINER_HOME is not defined.
        :param logger: a Logger object to use. By default a very simple logger is used, with INFO level, logging to stdout.
        :param loglevel: the loglevel, as an int value. Common values are defined in the standard logging module. Only used, if logger is not defined.
        :param rm_stdout: the output stream to redirect the output of underlying Studio launches. By default, the output is directed to the logger associated with this connector. Log records from Studio are labeled with new element 'key'='studio', while the logs from Python with 'key'='python'.
        :param override_python_binary: if set to True, any Execute Python operator in a process called via run_process will use the current Python interpreter instead of the one configured in the Studio installation, except if the process explicitly defines the Python binary. Default value is False.
        """
        super(Studio, self).__init__(logger, loglevel)
        if studio_home is not None:
            self.studio_home = studio_home
        elif os.getenv("RAPIDMINER_HOME") is not None and not os.getenv("RAPIDMINER_HOME") == "":
            self.studio_home = os.getenv("RAPIDMINER_HOME")
        else:
            # OS specific default home exists
            self.studio_home = Studio.__get_default_rmhome()
        self.__rm_stdout__ = rm_stdout
        self.override_python_binary = override_python_binary
        self.__last_exception_msg__ = {} # ensures proper multithreading: this maps last exception message for every thread
        self.__last_exit_code__ = {} # ensures proper multithreading: this maps last exit code for every thread
        self.__rapidminer_version = {}
        if not os.path.isdir(self.studio_home):
            raise StudioException("Specified or default RapidMiner Home does not exist: " + self.studio_home)
        scripts_dir = os.path.join(self.studio_home, self.__SCRIPTS_SUBDIR)
        if not os.path.isdir(scripts_dir):
            raise StudioException("Specified or default RapidMiner Home does not contain required scripts: " + scripts_dir)

####################
# Public functions #
####################

    def read_resource(self, path):
        """
        Reads one or more resources from the given repository locations / files

        Arguments:
        :param path: the path(s) to the resource(s). Multiple paths can be specified as list or tuple. A path can be a string, in that case it is interpreted as a repository location. By using rapidminer.File and rapidminer.RepositoryLocation objects, you can specify explicitly whether you want to use local files or RapidMiner repository entries.
        :return: the resource(s) as pandas DataFrame(s), a pickle-able python object(s) or a file-like object(s). If multiple inputs are specified, the same number of inputs will be returned, as a tuple of objects.
         """
        if not ((isinstance(path, tuple) or isinstance(path, list))):
            path = [path]
            single_input = True
        else:
            if len(path) == 0:
                return None
            single_input = False
        output_dirs = [tempfile.mkdtemp(prefix=self.__TMP_OUTPUT_DIR_PREFIX) for _ in path]
        try:
            self.__run_rapidminer(input_files=list(path), output_files=[File(output_dir) for output_dir in output_dirs], command_type="READ_RESOURCE")
            output_files = []
            for output_dir in output_dirs:
                csv_files = glob.glob(output_dir + "/*.csv-encoded")
                if (len(csv_files) == 1):
                    output_files.append(csv_files[0])
                else:
                    output_files.append(glob.glob(output_dir + "/*")[0])
            result = tuple(self.__deserialize_from_file(output_file) for output_file in output_files)
            if single_input:
                return result[0]
            else:
                return result
        finally:
            for dir in output_dirs:
                shutil.rmtree(dir, ignore_errors=True)

    def write_resource(self, resource, path):
        """
        Writes pandas DataFrame(s) or other objects to RapidMiner repository locations / regular files.

        Arguments:
        :param resource: resource(s) to save: a resource can be a pandas DataFrame, a pickle-able python object or a file-like object. Multiple DataFrames or other objects can be specified as list or tuple.
        :param path: the target path(s) for the resource(s). The same number of path values are required as the number of resources. A path can be a string, in that case it is interpreted as a repository location. By using rapidminer.File and rapidminer.RepositoryLocation objects, you can specify explicitly whether you want to use local files or RapidMiner repository entries.
        """
        if not ((isinstance(resource, tuple) or isinstance(resource, list))):
            resource = [resource]
        if not ((isinstance(path, tuple) or isinstance(path, list))):
            path = [path]

        if len(resource) != len(path):
            raise ValueError("Resource and path must contain the same number of values.")
        input_dirs = [tempfile.mkdtemp(prefix=self.__TMP_INPUT_DIR_PREFIX) for _ in resource]
        try:
            input_files = [self.__serialize_to_file(obj, os.path.join(dir, "input0")) for (dir, obj) in zip(input_dirs, resource)]
            self.__run_rapidminer(input_files=[File(f) for f in input_files], output_files=path, command_type="WRITE_RESOURCE")
        finally:
            for input_dir in input_dirs:
                shutil.rmtree(input_dir, ignore_errors=True)

    def run_process(self, path, inputs=[], macros={}, operator=None):
        """
        Runs a RapidMiner process and returns the result(s).

        Arguments:
        :param path: path to the RapidMiner process in a repository or an .rmp file. It can be a string, in that case it is interpreted as a repository location. It can also be a rapidminer.File or rapidminer.RepositoryLocation object representing a local file or a RapidMiner repository entry, respectively.
        :param inputs: inputs used by the RapidMiner process, an input can be a pandas DataFrame, a pickle-able python object or a file-like object.
        :param macros: optional dict that sets the macros in the process context according to the key-value pairs, e.g. macros={"macro1": "value1", "macro2": "value2"}
        :param operator: the name of the RapidMiner operator of the process to execute. If None (default) the whole process is executed.
        :return: the results of the RapidMiner process. It can be None, or a single object, or a tuple. One result may be a pandas DataFrame, a pickle-able python object or a file-like object.
        """
        if inputs is not None and not (isinstance(inputs, tuple) or isinstance(inputs, list)):
            inputs = [inputs]
        output_dir = tempfile.mkdtemp(prefix=self.__TMP_OUTPUT_DIR_PREFIX)
        remove_dirs = [output_dir]
        try:
            input_files = []
            if inputs:
                input_dir = tempfile.mkdtemp(prefix=self.__TMP_INPUT_DIR_PREFIX)
                remove_dirs.append(input_dir)
                for i in range(len(inputs)):
                    input_files.append(File(self.__serialize_to_file(inputs[i], os.path.join(input_dir, "input" + str(i)))))
            return self.__run_process_with_output_dir(path, input_files, operator, output_dir, macros, command_type="RUN_PROCESS")
        finally:
            for dir in remove_dirs:
                shutil.rmtree(dir, ignore_errors=True)

#####################
# Private functions #
#####################

    def __get_default_rmhome():
        if platform.system() == "Windows":
            return "C:/Program Files/RapidMiner Studio"
        elif platform.system() == "Darwin":
            return "/Applications/RapidMiner Studio.app/Contents/Resources/RapidMiner-Studio"
        else:
            # no default on other systems
            return os.getcwd()

    def __extract_log_level(self, msg):
        # LogLevels: https://docs.python.org/2/library/logging.html#logging-levels
        if msg.startswith("FINEST: "):
            lglevel = logging.DEBUG
            msg = msg[8:]
        elif msg.startswith("FINER: "):
            lglevel = logging.DEBUG
            msg = msg[7:]
        elif msg.startswith("DEBUG: "):
            lglevel = logging.DEBUG
            msg = msg[7:]
        elif msg.startswith("CONFIG: "):
            lglevel = logging.DEBUG
            msg = msg[8:]
        elif msg.startswith("INFO: "):
            lglevel = logging.INFO
            msg = msg[6:]
        elif msg.startswith("WARNING: "):
            lglevel = logging.WARNING
            msg = msg[9:]
        elif msg.startswith("SEVERE: "):
            lglevel = logging.ERROR
            msg = msg[8:]
        elif msg.startswith(self.__RAPIDMINER_ERROR_MSG_FIRST_LINE):
            lglevel = -1
        elif msg.startswith(self.__RAPIDMINER_ERROR_MSG):
            lglevel = logging.ERROR
            msg = msg[len(self.__RAPIDMINER_ERROR_MSG):]
        elif msg.startswith(self.___EXIT_CODE_MSG):
            lglevel = -1
        else:
            lglevel = logging.INFO
        return (msg, lglevel)

    def __update_version(self, msg, threadid):
        if msg.startswith(self.__VERSION_MSG):
            msg = msg[len(self.__VERSION_MSG):]
            self.__rapidminer_version[threadid] = msg

    def __update_error_and_exit_code_fields(self, msg, threadid):
        if msg.startswith(self.__RAPIDMINER_ERROR_MSG_FIRST_LINE):
            msg = msg[len(self.__RAPIDMINER_ERROR_MSG_FIRST_LINE):]
            self.__last_exception_msg__[threadid] = msg
        elif msg.startswith(self.___EXIT_CODE_MSG):
            try:
                self.__last_exit_code__[threadid] = int(msg[10:])
            except ValueError:
                self.__last_exit_code__[threadid] = 0

    def __print_to_console(self, process, close_process_stdout=False, threadid = -1):
        for line in iter(process.stdout.readline, b''):
            try:
                msg = line.decode(encoding=__STDOUT_ENCODING__, errors='ignore')
                self.__update_version(msg, threadid)
                self.__update_error_and_exit_code_fields(msg, threadid)
                if self.__rm_stdout__ is not None:
                    self.__rm_stdout__.write(msg)
                else:
                    (msg, lglevel) = self.__extract_log_level(msg)
                    self.log(msg, level=lglevel, source="studio")
            except UnicodeEncodeError:
                self.log("<could not decode row>", level=logging.DEBUG, source="studio")
        if close_process_stdout:
            process.stdout.close()

    def __start_printer_thread(self, process):
        t = Thread(target = self.__print_to_console, args=(process, False, threading.currentThread().ident))
        t.daemon = True
        t.start()

    def __quote_params(self, param, prefix=""):
        if platform.system() == "Windows":
            return prefix + param
        else:
            return '\"' + prefix + param + '\"'

    def __encode_params(self, param):
        encoded = ""
        for character in param:
            if ord(character) < 128:
                if character == "\\":
                    # replace \ with \\
                    encoded = encoded + "\\\\"
                elif character == "\"":
                    # replace " with \a
                    encoded = encoded + "\\a"
                else:
                    encoded = encoded + character
            else:
                #replace non ASCII with \nxHHHH, where n is the length og HHHH, HHHH is the hex code
                code = hex(ord(character))[2:]
                encoded = encoded + "\\" + str(len(code)) + "x" + code
        return encoded

    def __append_param(self, params, param, prefix):
        params.append(self.__quote_params(self.__encode_params(param), prefix=prefix))

    def __get_script_suffix(self):
        if platform.system() == "Windows":
            return ".bat"
        elif platform.system() == "Darwin":
            return "-osx.sh"
        else:
            return ".sh"

    def __needs_temp_dir(self, input_file):
        '''
        Returns true, if the given input file will need a temporary directory on the RapidMiner side.

        :param input_file: the file to be inspected. Files with .fo extension needs a temp dir. (file-object)
        :return: true, if the given input file will need a temporary directory on the RapidMiner side.
        '''
        if isinstance(input_file, Resource):
            input_file = input_file.to_string()
        return input_file.endswith(".fo")

    # TODO refactor this method to reduce its cognitive complexity from 26 to the allowed 15...
    def __run_rapidminer(self, process=None, input_files=[], output_files=[], output_dir=None, macros={}, operator=None, command_type=None):
        kwargs = {"stdout": subprocess.PIPE,
                  "stderr": subprocess.STDOUT,
                  "bufsize": 10}
        params = []
        params.append(os.path.join(self.studio_home, self.__SCRIPTS_SUBDIR, "rapidminer-batch" + self.__get_script_suffix()))
        self.__append_param(params, "rmx_python_scripting:com.rapidminer.extension.pythonscripting.launcher.ExtendedCmdLauncher", prefix="-C")
        self.__append_param(params, __version__, prefix="-V")
        if (process is not None):
            if not isinstance(process, Resource):
                process = RepositoryLocation(name=process)
            self.__append_param(params, process.to_string(), prefix="-P")
        for input_file in input_files:
            if not isinstance(input_file, Resource):
                input_file = RepositoryLocation(name=input_file)
            self.__append_param(params, input_file.to_string(), prefix="-I")
        for output_file in output_files:
            if not isinstance(output_file, Resource):
                output_file = RepositoryLocation(name=output_file)
            self.__append_param(params, output_file.to_string(), prefix="-O")
        if output_dir is not None:
            self.__append_param(params, output_dir, prefix="-D")
        if operator is not None:
            self.__append_param(params, operator, prefix="-N")
        if len(macros) > 0:
            for key in macros:
                self.__append_param(params, str(key) + "=" + str(macros[key]), prefix="-M")
        if any(self.__needs_temp_dir(input) for input in input_files):
            temp_dir = tempfile.mkdtemp(prefix=self.__TMP_OUTPUT_DIR_PREFIX)
            self.__append_param(params, temp_dir, prefix="-T")
        else:
            temp_dir = None
        if self.override_python_binary:
            self.__append_param(params, sys.executable, prefix="-B")
        if command_type is not None:
            self.__append_param(params, command_type, prefix="-A")
        threadid = threading.currentThread().ident
        if threadid in self.__last_exit_code__:
            del self.__last_exit_code__[threadid]
        if threadid in self.__last_exception_msg__:
            del self.__last_exception_msg__[threadid]
        if threadid in self.__rapidminer_version:
            del self.__rapidminer_version[threadid]
        try:
            p = subprocess.Popen(params, **kwargs)
            try:
                self.__start_printer_thread(p)
                p.wait()
                if not threadid in self.__rapidminer_version or not Version(self.__rapidminer_version[threadid]).is_at_least(Version("9.5.0")):
                    raise VersionException("RapidMiner Studio", "9.5.0 or newer")
                if not threadid in self.__last_exit_code__ or self.__last_exit_code__[threadid] != 0:
                    if threadid in self.__last_exception_msg__:
                        raise StudioException("Error while executing studio: " + self.__last_exception_msg__[threadid])
                    elif threadid in self.__last_exit_code__:
                        raise StudioException("Error while executing studio - unknown error. (error code: " + str(self.__last_exit_code__) + ")")
                    else:
                        raise StudioException("Error while executing studio - unknown error.")
            finally:
                p.stdout.close()
        finally:
            if temp_dir is not None:
                shutil.rmtree(temp_dir, ignore_errors=True)

    def __run_process_with_output_dir(self, path, input_files, operator, output_dir, macros, command_type):
        self.__run_rapidminer(process=path, input_files=input_files, output_dir=output_dir, macros=macros, operator=operator, command_type=command_type)
        outputs = glob.glob(os.path.join(output_dir, "*.*"))
        outputs.sort()
        result = []
        for output in outputs:
            if not output.endswith(".pmd-encoded"):
                result.append(self.__deserialize_from_file(output))
        if len(result) == 0:
            return None
        elif len(result) == 1:
            return result[0]
        return tuple(result)

    def __serialize_to_file(self, object, basename):
        """
        Serializes a python object to the appropriate file.

        :param object, a python object.
        :param basename: the base filename, without extension.
        :return:
        """
        if isinstance(object, pandas.DataFrame):
            with __open__(basename + ".csv-encoded", "w") as output_csv:
                with __open__(basename + ".pmd-encoded", "w") as output_pmd:
                    write_example_set(self._copy_dataframe(object), output_csv, output_pmd)
            return basename + ".csv-encoded"
        else:
            # try to write out as a file like object first
            try:
                with open(basename + ".fo", "w", encoding=object.encoding) as outf:
                    shutil.copyfileobj(object, outf)
                return basename + ".fo"
            except AttributeError:
                try:
                    with open(basename + ".fo", "wb") as outf:
                        shutil.copyfileobj(object, outf)
                    return basename + ".fo"
                except AttributeError:
                    shutil.rmtree(basename + ".fo", ignore_errors=True)
                    with open(basename + ".bin", 'wb') as dump_file:
                        pickle.dump(object, dump_file)
                    return basename + ".bin"

    def __deserialize_from_file(self, filename):
        """
        Reads the given file. The actual method depends on the file extension

        :param filename: name of the file
        :return: an arbitrary python object (DataFrame, file object or any other python type pickled out)
        """
        (path, extension) = os.path.splitext(filename)
        if extension=='.csv-encoded':
            with __open__(path + ".csv-encoded", 'r') as input_csv:
                with __open__(path + ".pmd-encoded", 'r') as input_pmd:
                    return read_example_set(input_csv, input_pmd)
        elif extension=='.bin':
            with open(filename, 'rb') as f:
                try:
                    return pickle.load(f)
                except Exception as exc:
                    raise GeneralException("Error while trying to load pickled object:" + str(exc))
        elif extension=='.fo':
            with open(filename, 'rb') as f:
                return io.BytesIO(f.read()) # reads the file to memory
        else:
            raise ValueError("Cannot handle files with '" + str(extension) + "' extension.")



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import pandas
import json
import sys
import numpy as np
import base64
import warnings
import datetime

try:
    from .. import __version__
except:
    __version__ = "0.0.0" # replaced by extension code

# instance(x, basestring) works with Python 2 and 3 this way
try:
    basestring
except NameError:
    basestring = str

class DateConversionError(ValueError):
    def __init__(self, msg):
        super(DateConversionError, self).__init__(msg)

# returns the error raised for empty inputs by pandas.read_csv (pandas version dependent)
def __get_empty_data_error():
    try:
        return pandas.errors.EmptyDataError
    except:
        try:
            return pandas.parser.CParserError, ValueError
        except:
            return None

def __open_file_python2(file, mode):
    return open(file, mode)

def __open_file_python3(file, mode):
    return open(file, mode, encoding="utf-8")

def __write_file_python2(file, object):
    if type(object) != unicode:
        if type(object) != str:
            object = str(object)
        object = unicode(object, "utf-8")
    file.write(object.encode("utf-8"))

def __write_file_python3(file, object):
    file.write(str(object))

def read_file(file):
    return file.read()

def __b64encode_python2(str):
    return base64.b64encode(str)

def __b64encode_python3(str):
    return base64.b64encode(str.encode("utf-8")).decode("utf-8")

def __b64decode_python2(str):
    return base64.b64decode(str)

def __b64decode_python3(str):
    return base64.b64decode(str.encode("utf-8")).decode("utf-8")

if sys.version_info >= (3, 0):
    open_file = __open_file_python3
    write_file = __write_file_python3
    b64encode = __b64encode_python3
    b64decode = __b64decode_python3
else:
    open_file = __open_file_python2
    write_file = __write_file_python2
    b64encode = __b64encode_python2
    b64decode = __b64decode_python2

def is_file_object(entry):
    try:
        if entry.closed:
            open(entry)
            close(entry)
        return True
    except Exception:
        return False

# writes string to error file
def write_to_error_log(full_message):
    with open_file("rapidminer_error.log", "w") as error_file:
        write_file(error_file, full_message)

# transforms meta data list to a dictionary
def transform_metadata(list):
    dict = {}
    for element in list:
        name, (type, role) = next(iter(element.items()))
        name = __handle_unicode(name)
        type = __handle_unicode(type)
        role = __handle_unicode(role)

        # "attribute" role is handled differently due to backward compatibility
        if role == "attribute":
            dict[name] = (type, None)
        else:
            dict[name] = (type, role)
    return dict

# reads the example set from the files into a pandas.DataFrame.
def read_example_set(input_csv, input_pmd):
    def read_date(epoch):
        if epoch != "null" and epoch is not np.nan and epoch == epoch:
            # because of some older pandas versions, we need to read in "ns"
            try:
                return pandas.to_datetime(int(epoch)*1000, unit='ns')
            except OverflowError:
                raise Exception("Overflow error occurred during reading date columns. Python pandas can only handle dates between " + str(datetime.date(1677,9,21)) + " and " + str(datetime.date(2262,4,11)) + ".\n\nIf you want to use dates outside this range, please convert your data to integer in RapidMiner with 'Date to Numerical' operator, and convert back integer values in Python pandas manually in your code.")
        else:
            return None

    # metadata
    metadata = json.loads(read_file(input_pmd))["metadata"]
    dtype = {}
    rename_map = {}
    conv = {}
    date_cols = []
    index = 0
    for attr_info in metadata:
        attr, (rm_type, _) = next(iter(attr_info.items()))
        attr = __handle_unicode(attr)
        rm_type = __handle_unicode(rm_type)
        if rm_type in ["real", "numeric"]:
            dtype[index] = "float64"
        elif __nominal_meta_type(rm_type):
            conv[index] = lambda x: b64decode(x) if x != "null" else np.nan
        elif __date_meta_type(rm_type):
            # "datetime64[ms]" is not allowed
            dtype[index] = "object"
            date_cols.append(index)
        elif rm_type != "integer":
            # "int64" may fail because of missings, not setting type in this case: it will either be float64 (missings) or int64
            print("type in '" + rm_type + "' is not a valid rapidminer type")
            write_to_error_log("type in '" + rm_type + "' is not a valid rapidminer type")
            sys.exit(65)
        else:
            dtype[index] = "int64"
        rename_map[index] = attr
        index = index + 1
    # data
    with warnings.catch_warnings():
        # potential irrelevant warnings on nominal columns
        try:
            warnings.simplefilter("ignore", pandas.errors.DtypeWarning)
        except:
            # older Pandas version does not have pandas.errors
            pass
        try:
            df = pandas.read_csv(input_csv,
                                 dtype={i:dtype[i] for i in dtype.keys() if dtype[i] != "int64"},
                                 encoding="utf-8",
                                 header=None,
                                 converters=conv,
                                 parse_dates=date_cols,
                                 date_parser=lambda epoch: read_date(epoch))
            df.rename(columns=rename_map, inplace=True)
        except __get_empty_data_error():
            df = pandas.DataFrame([["obj" for _ in range(len(rename_map))]], columns=[rename_map[i] for i in range(len(rename_map))])
            df = df.iloc[0:0]
            if len(dtype) > 0:
                dtype_map = {rename_map[i]:dtype[i] for i in dtype.keys()}
                for column in dtype_map.keys():
                    df[column] = df[column].astype(dtype_map[column])
    set_metadata_without_warning(df, transform_metadata(metadata))
    return df

# uses the meta data from rm_metadata attribute if present
# otherwise deduces the type from the data and sets no special role
def get_metadata(data, original_names):
    metadata = []

    # check if rm_metadata attribute is present and a dictionary
    try:
        if hasattr(data, "rm_metadata") and isinstance(data.rm_metadata,dict):
            meta_isdict = True
        elif hasattr(data, "rm_metadata"):
            meta_isdict = False
            if data.rm_metadata is not None:
                print("Warning: rm_metadata must be a dictionary")
        else:
            meta_isdict = False
    except:
        meta_isdict = False

    for name in data.columns.values:
        try:
            original_name = original_names[name] if name in original_names else name
            if hasattr(data, "rm_metadata"):
                meta = data.rm_metadata[original_name]
            else:
                meta = None
            #take entries only if tuple of length 2
            if isinstance(meta,tuple) and len(meta)==2 and meta_isdict:
                meta_type, meta_role = meta
            else:
                if meta_isdict and meta is not None:
                    print("Warning: rm_metadata["+str(original_name)+"] must be a tuple of length 2, e.g. data.rm_metadata['column1']=('binominal','label')")
                # if the format of metadata is not correct, still try to figure out meta type and role
                if isinstance(meta, basestring) and len(meta) > 0:
                    meta_type = meta
                    meta_role = None
                elif (isinstance(meta, tuple) or isinstance(meta, list)) and len(meta) > 0:
                    meta_type = meta[0]
                    if len(meta) > 1:
                        meta_role = meta[1]
                    else:
                        meta_role = None
                else:
                    meta_type = None
                    meta_role = None
            meta_type = __handle_unicode(meta_type)
            meta_role = __handle_unicode(meta_role)
        except Exception as e:
            print(e)
            meta_type = None
            meta_role = None

        if meta_role is None:
            meta_role = 'attribute'
        #choose type by dtype of the column
        if meta_type is None or not __valid_meta_type(meta_type):
            kind_char = data.dtypes[name].kind
            if kind_char in ('i','u'):
                meta_type = 'integer'
            elif kind_char in ('f'):
                meta_type = 'real'
            elif kind_char in ('M'):
                meta_type = 'date_time'
            elif kind_char in ('b'):
                meta_type = 'binominal'
            else:
                meta_type = 'polynominal'
        # double quote and backslash characters are escaped automatically in the name key
        metadata.append({name : (meta_type,meta_role)})
    return metadata

# if name has a string representation containing non-ascii symbols in python 2,
# for example if name is a python 2 unicode with umlauts, then str(name) results in exception;
# in this case it is in particular not empty and contains more than only digits
def isstringable(name):
    try:
        str(name)
        return True
    except:
        return False

def is_invalid_name(name):
    return isstringable(name) and ((not str(name)) or str(name).isdigit())

def rename_columns(dataframe):
    original_names = {}
    # column name must not be empty or a number
    if any(is_invalid_name(name) for name in dataframe.columns.values):
        new_columns = []
        for name in dataframe.columns.values:
            new_name = name
            if is_invalid_name(name):
                new_name = 'att'+str(name)
                original_names[new_name] = name
            new_columns.append(new_name)
        dataframe.columns = new_columns
    return original_names

# base64 encode all nominals and convert dates to epoch
def convert_to_output_format(df, metadata, hdf5_compliant_date_and_time_conversion):
    def b64(x):
        if isinstance(x, basestring):
            return b64encode(x)
        # avoid lists, numpy array, etc. in isnull check
        elif not hasattr(x, "__len__") and pandas.isnull(x):
            return "null"
        return b64encode(str(x))

    if (hasattr(df, "rm_converted_for_writing")) and df.rm_converted_for_writing:
        return

    for m in metadata:
        # there is just one column in one list element
        for name in m:
            meta_type = m[name][0]
            if __date_meta_type(meta_type):
                index = np.logical_not(df[name].isnull())
                if not any(index):
                    # we have only null values
                    continue
                try:
                    if all(index):
                        # // operator is necessary to keep type int in all python, pandas versions
                        df[name] = df[name].astype("int64")//1000
                    else:
                        # workaround for pandas error in some older pandas version
                        if type(index) == pandas.Series:
                            try:
                                # this will always raise errors with older pandas versions
                                index[0:0].dropna()
                            except TypeError:
                                index = index.tolist()
                        df.loc[index, name] = df.loc[index,name].astype("int64")
                        # check if conversion succeeded (for pandas==0.17.1 for example it is not working)
                        sample = df.loc[index, name].iloc[0]
                        if __is_integer_number(sample):
                            # // operator is necessary to keep type int in all python, pandas versions
                            df.loc[index, name] = df.loc[index,name]//1000
                        else:
                            df[name] = df[name].astype("object")
                            df.loc[index, name] = ((df.loc[index, name] - pandas.to_datetime(0, unit="ns")).dt.total_seconds()*1e6).round().astype("int64")
                    if hdf5_compliant_date_and_time_conversion:
                        if meta_type == "time":
                            df[name] = df[name] % 86400000000
                        elif meta_type == "date":
                            df[name] = df[name] // 1e6 * 1e6
                except ValueError:
                    raise DateConversionError("Error while serializing dataframe: some values in column '" + str(name) + "' are not valid dates.")
            elif __nominal_meta_type(meta_type):
                df[name] = df[name].apply(b64)
    df.rm_converted_for_writing = True

# writes the data and metadata to files
# SIDE EFFECT: the method may modify the original data
def write_example_set(data, output_csv, output_pmd, hdf5_compliant_date_and_time_conversion=False):
    original_names = rename_columns(data)
    # metadata
    metadata = get_metadata(data, original_names)
    write_file(output_pmd, "{\n")
    write_file(output_pmd, "  \"source\": \"RapidMiner Python Scripting Extension and Library\",\n")
    write_file(output_pmd, "  \"module\": \"Python\",\n")
    write_file(output_pmd, "  \"version\": \"" + __version__ + "\",\n")
    write_file(output_pmd, "  \"metadata\":\n" + json.dumps(metadata))
    write_file(output_pmd, "\n}")
    # data
    convert_to_output_format(data, metadata, hdf5_compliant_date_and_time_conversion)
    data.to_csv(output_csv, encoding="utf-8", header=False, index=False)

def set_metadata_without_warning(df, metadata):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        df.rm_metadata = metadata

# based on Ontology
def __valid_meta_type(meta_type):
    return meta_type in [None, "integer", "real", "numeric"] or __nominal_meta_type(meta_type) or __date_meta_type(meta_type)

def __nominal_meta_type(meta_type):
    return meta_type in ["nominal", "binominal", "polynominal", "text", "file_path"]

def __date_meta_type(meta_type):
    return meta_type in ["date_time", "date", "time"]

# ensures python2 compatibility
def __handle_unicode(value):
    if sys.version_info.major == 2:
        if type(value) == unicode:
            return value.encode("utf-8")
    return value

def __is_integer_number(vut):
    try:
        vut + 1
        return True
    except:
        return False



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import json
import logging
import sys
import threading
import datetime
import pandas as pd
from collections import OrderedDict
from .serdeutils import set_metadata_without_warning

class Connector(object):
    """
    Base class for interacting with RapidMiner. The subclasses of this class should be used.
    """
    __date_format_pattern = "%Y-%m-%d %H:%M:%S.%f"
    __id_counter__ = 0
    __lock__ = threading.Lock()

    def __init__(self, logger=None, loglevel=logging.INFO):
        """
        Arguments:
        :param logger: a Logger object to use. By default a very simple logger is used, with INFO level, logging to
                        stdout.
        :param loglevel: the loglevel, as an int value. Common values are defined in the standard logging module. Only
                        used if logger is not defined.
        """
        Connector.__lock__.acquire()
        try:
            self.__id__ = Connector.__id_counter__
            Connector.__id_counter__ = Connector.__id_counter__ + 1
        finally:
            Connector.__lock__.release()
        if logger != None:
            self.logger = logger
        else:
            formatter=logging.Formatter("%(asctime)s [%(levelname)s -- %(source)s]: %(message)s")
            syslog=logging.StreamHandler(sys.stdout)
            syslog.setFormatter(formatter)
            self.logger = logging.getLogger(self.__class__.__name__ + "@" + str(self.__id__))
            self.logger.setLevel(loglevel)
            self.logger.addHandler(syslog)

    def log(self, msg, level=logging.INFO, source="python"):
        """
        Logs a message with the defined log level.

        Arguments:
        :param msg: the message to log.
        :param level: the log level as an integer.
        :param source: source of the message, default is 'python'.
        :return:
        """
        self.logger.log(msg=msg.strip(), level=level, extra={"source": source})

    def read_resource(self, path):
        """
        Reads one or more resources from the given locations.

        Arguments:
        :param path: the path(s) to the resource(s). Multiple paths can be specified as list or tuple.
        :return: the resource(s) read from the path(s). If multiple inputs are specified, the same number of inputs will be returned, as a tuple of objects.
        """
        raise NotImplementedError("Method not implemented in base class.")

    def write_resource(self, resource, path):
        """
        Writes the pandas DataFrame to RapidMiner repository/regular file.

        Arguments:
        :param resource: resource(s) to save. Multiple DataFrames or other objects can be specified as list or tuple.
        :param path: the target path(s) for the resource(s). The same number of path values are required as the number of resources.
        """
        raise NotImplementedError("Method not implemented in base class.")


    def run_process(self, path, inputs=[], macros={}):
        """
        Runs a RapidMiner process and returns the result(s).

        Arguments:
        :param path: path to the RapidMiner process.
        :param inputs: inputs used by the RapidMiner process, as a list of input objects or a single input object.
        :param macros: optional dict that sets the macros in the process context according to the key-value pairs, e.g. macros={"macro1": "value1", "macro2": "value2"}
        :return: the results of the RapidMiner process. It can be None, or a single object, or a tuple.
        """
        raise NotImplementedError("Method not implemented in base class.")

    def _can_convert_to_str(self, value):
        """
        Tests, if the given value can be converted to a string representation.

        Taken from the legacy wrapper.py code (isstringable).

        :param value: value to test.
        :return: True, if value can be converted to string, False otherwise.
        """
        try:
            str(value)
            return True
        except:
            return False

    def _rename_invalid_columns(self, columns):
        """
        Renames the invalid column names. Column names must be not empty and not a single number.

        Taken -- with some modification -- from the legacy wrapper.py code (checkColumNames).

        :param columns: list of DataFrame columns.
        :return:
        """
        if any(self._can_convert_to_str(value) and ((not str(value)) or str(value).isdigit()) for value in columns):
            return ['att'+str(value) if (self._can_convert_to_str(value) and ((not str(value)) or str(value).isdigit())) else str(value) for value in columns]
        else:
            return columns

    def _copy_dataframe(self, df):
        """
        Copies the dataframe, together with rm_metadata
        :param df:
        :return:
        """
        copied = df.copy()
        if hasattr(df, "rm_metadata"):
            set_metadata_without_warning(copied, df.rm_metadata)
        return copied



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#

import os
import glob
import json
import zipfile
from collections import OrderedDict
from tink import aead
from tink import cleartext_keyset_handle
from tink import JsonKeysetReader
import base64
from .utilities import ProjectException
from .utilities import ServerException
from .utilities import _is_docker_based_deployment
from .resources import ProjectLocation


class Connections():
    """
    Class for using connections from a Project or Repository.
    """
    
    _CONNECTIONS_SUBDIR = "Connections"
    _CONNECTIONS_EXTENSION = ".conninfo"
    
    def __init__(self, path=".", server=None, project_name=None, show_parameter_groups=False, macros=None):
        """
        Initializes a reference to a locally cloned project or to an AI Hub repository. You either need to clone a project from AI Hub first (e.g. via git commands), or point to an AI Hub repository to be able to use the methods of this class.
        
        :param path: path to the local project repository root folder. It can be a relative path from the current working directory or an absolute path. The default value points to the working directory. Set it to None to read connections from AI Hub repository (server parameter needs to be set in this case)
        :param server: Server object; required when values are encrypted or when values are injected from the AI Hub Vault
        :param project_name: name of the project to which these connections belong; if not specified, it is set to the parent directory name of the specified path parameter. Ignored if AI Hub repository connections are used
        :param show_parameter_groups: when set to True (default if False), parameter keys include the group as well in group.key format; otherwise, group is only added if there is a name collision among the keys
        :param macros: a dictionary or method defining the values for the macro injected parameters. If a connection has a prefix defined for macro keys, the prefix is added before the key. If a method is used here, it should accept connection_name and macro_name parameters, and should return the macro value for those
        """
        if not path is None:
            if not os.path.exists(path):
                raise ProjectException("Specified path does not exist: '%s'. Please make sure you have a local copy of the project at the specified path." % path)
            path = os.path.abspath(path)
            # if not specified, derive project name from the directory name
            if not project_name:
                project_name = os.path.basename(path)
            path = os.path.join(path, Connections._CONNECTIONS_SUBDIR)
            if not os.path.exists(path):
                raise ProjectException("Connections directory does not exist at the specified path '%s'. Please make sure you have a local copy of the project." % (os.path.dirname(path)))
        self.path = path
        aead.register()

        self.server = server
        self.project_name = project_name
        self._show_parameter_groups = show_parameter_groups
        self.macros = macros
        self.__cached_project_primitive = None
        self.__list = []
        
        # load connections either from legacy AI Hub repository or from project path
        if self.path is None:
            self._check_server("You must either provide a project path ('path' parameter) or a Server object ('server' parameter).")
            conn_json = self.server._get_connections_info()
            for c in conn_json:
                self.__list.append(Connection(c["location"], OrderedDict(server._read_connection_info(c["location"])), self))
        else:
            conn_files = glob.glob(os.path.join(self.path, "*" + Connections._CONNECTIONS_EXTENSION))
            for z in conn_files:
                with zipfile.ZipFile(z) as zf:
                    with zf.open("Config") as f:
                        self.__list.append(Connection(os.path.join(Connections._CONNECTIONS_SUBDIR, os.path.basename(z)),
                                                      json.loads(f.read(), object_pairs_hook=OrderedDict), self))
        self.__list.sort(key=lambda c: c.name)

    def _get_project_primitive(self):
        # project primitive is not expected to change - initialized once, when first needed
        if self.__cached_project_primitive == None:
            response = self.server._get_project_info(self.project_name)
            content = json.dumps(response["secret"])
            reader = JsonKeysetReader(content)
            keyset_handle = cleartext_keyset_handle.read(reader)
            self.__cached_project_primitive = keyset_handle.primitive(aead.Aead)
        return self.__cached_project_primitive

    def _check_server(self, error_message):
        if not self.server:
            if _is_docker_based_deployment():
                self.server = get_server()
            else:
                raise ServerException(error_message)
    
    def __iter__(self):
        return iter(self.__list)

    def __getitem__(self, item):
        try:
            return self.__list[int(item)]
        except ValueError:
            return next(x for x in self.__list if x.name == item)
        
    def __str__(self):
        return "Connections(%s)" % [x.name for x in self.__list]

    def __repr__(self):
        return repr("Connections(%s)" % [x.name for x in self.__list])


class Connection():
    """
    Class that represents a single connection.
    """

    def __init__(self, location, config, connections):
        """
        Initializes a connection instance based on a location and a config dictionary.

        :param location: path to the connection file in the project or in the AI Hub repository
        :param config: dictionary containing all the connection details; can be directly created from the JSON content of the connection
        :param connections: reference to the parent Connections object
        """
        self.__location = location
        self.__config = config
        self.__connections = connections
        self.__name = config["name"]
        self.__type = config["type"]
        self.__init_constant_values()

    @property
    def config(self):
        """
        The raw internal JSON configuration of the connection.
        """
        return self.__config

    @property
    def name(self):
        """
        The name of the connection.
        """
        return self.__name

    @property
    def type(self):
        """
        Type of the connection, i.e. what kind of system or service it connects to.
        """
        return self.__type

    @property
    def values(self):
        """
        Dictionary with all the connection fields. Injected and encrypted fields are all handled transparently provided that the appropriate information is accessible. Otherwise, an error is thrown for the first problem encountered. Note that encrypted values from an AI Hub repository are not available (values will be None) - use AI Hub Vault in this case. When accessing multiple fields, you may want to first create a copy from the returned dictionary to avoid retrieving values stored in AI Hub vault or decrypting values multiple times (all such values are refreshed in each call).
        """
        self.__refresh_dynamic_values()
        return self.__values

    # Accessing some of the typical fields easily
    @property
    def user(self):
        """
        Quick way to access the first field that probably contains a user name, e.g. the field name contains the string username.
        """
        return self.find_first("username", "user")

    @property
    def password(self):
        """
        Quick way to access the first field that probably contains a password, e.g. the field name contains the string password.
        """
        return self.find_first("password")
    
    def find_first(self, *words):
        """
        Returns the value of the first connection parameter that has one of the specified arguments in its key as a substring. Useful for looking up parameters if the key is not known accurately.
        
        :param words: arbitrary number of words to look for - the functions stops at the very first word for which it successfully finds a parameter that has the word in its key
        """
        values_cache = self.values
        for w in words:
            for key in values_cache:
                # don't search in groups
                if w.lower() in key.split(".")[-1].lower():
                    return values_cache[key]
        raise KeyError("Could not find any parameters having the words %s in their key." % str(words))

    def __init_constant_values(self):
        self.__values = OrderedDict()
        self.__lazy_loaded_params = []
        param_keys = set()
        for k in self.__config["keys"]:
            self.__init_constant_values_in_group(k,param_keys)

    def __init_constant_values_in_group(self, group_config, param_keys):
        group = group_config["group"]
        for p in group_config["parameters"]:
            # only add group if there would be a conflict in keys otherwise, or if flag variable is set to True
            if self.__connections._show_parameter_groups or p["name"] in param_keys:
                p["name"] = group + "." + p["name"]
            p["group"] = group
            param_keys.add(p["name"])
            if p["enabled"]:
                if ("injectorName" in p and p["injectorName"]) or p["encrypted"]:
                    self.__values[p["name"]] = None
                    self.__lazy_loaded_params.append(p)
                else:
                    self.__values[p["name"]] = p["value"] if "value" in p else None
    
    def __refresh_dynamic_values(self):
        if len(self.__lazy_loaded_params) < 1:
            # nothing to refresh
            return
        for n in self.__lazy_loaded_params:
            self.__refresh_dynamic_value(n)

    def __refresh_dynamic_value(self, parameter):
        self.__refresh_vault_cache()
        name = parameter["name"]
        value = parameter["value"] if "value" in parameter else None
        if "injectorName" in parameter and parameter["injectorName"]:
            value = self.__injected_value(parameter)
            # when injected, there is no encryption
        elif parameter["encrypted"]:
            value = self.__decrypt(value)
        self.__values[name] = value

    def __refresh_vault_cache(self):
        self.__connections._check_server("For accessing a value from the AI Hub Vault, you must provide a Server object.")
        if self.__connections.project_name is None:
            location = self.__location
        else:
            location = ProjectLocation(self.__connections.project_name, self.__location)
        self.__cached_vault_info = self.__connections.server._get_vault_info(location)
    
    def __decrypt(self, value):
        if not value:
            return None
        self.__connections._check_server("This connection has encrypted values. Decrypting them is only supported in case of AI Hub (Server) projects/repositories. For decrypting encrypted values, you must provide a Server object.")
        # Use key from Server
        res = self.__connections._get_project_primitive().decrypt(base64.b64decode(value), b'')
        return res.decode("utf-8")

    def __injected_value(self, parameter):
        injector = self.__injector(parameter["injectorName"])
        if injector:
            if injector["type"] == "remote_repository:rapidminer_vault":
                return self.__injected_vault_value(parameter)
            elif injector["type"] == "macro_value_provider":
                return self.__injected_macro_value(parameter, injector)
            else:
                raise ValueError("Unknown injector type %s" % injector["type"])
        return None

    def __injected_vault_value(self, parameter):
        # cut the group if <group>.<key> format is used
        name = parameter["name"].split(".")[-1]
        for v in self.__cached_vault_info:
            if v["parameter"]["name"] == name and v["parameter"]["key"]["group"] == parameter["group"]:
                return v["value"]
        raise ValueError("Parameter with key %s not found in Server Vault" % parameter["name"])

    def __injected_macro_value(self, parameter, injector):
        # cut the group if <group>.<key> format is used
        name = parameter["name"].split(".")[-1]
        if injector["parameters"][0]["value"]:
            # macro prefix is used
            name = str(injector["parameters"][0]["value"]) + "_" + str(name)
        macros = self.__connections.macros
        if macros:
            return self.__get_macro(macros, name)
        raise ValueError("Connection %s uses macros, but macros parameter is set to None." % self.__name)

    def __get_macro(self, macros, key):
        if callable(macros):
            try:
                return macros(self.__name, key)
            except Exception as e:
                raise ValueError("Cannot get macro key %s for connection %s. Reason: %s" % (key, self.__name, str(e)), e)
        else:
            try:
                return macros[key]
            except KeyError:
                raise ValueError("Macro for key %s is not defined in macros dictionary." % key)

    def __injector(self, injector_name):
        for p in self.__config["valueProviders"]:
            if p["name"] == injector_name:
                return p
        raise ValueError("Value provider for injector type %s not found" % injector_name)

    def __str__(self):
        return "Connection(%s)" % self.name

    def __repr__(self):
        return repr("Connection(%s)" % self.name)




#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
class Resource(object):
    def to_string(self):
        """
        Returns a string based representation of this resource, which is handled by Rapidminer Studio

        :return: String representation.
        """
        raise NotImplementedError("Method not implemented in base class.")


class File(Resource):
    def __init__(self, filename):
        if not isinstance(filename, str):
            raise ValueError("'filename' must be a string. (now: " + str(type(filename)) +")")
        self.filename = filename

    def to_string(self):
        return "file:" + self.filename


class RepositoryLocation(Resource):
    __SEP__ = "/"

    def __init__(self, parent=None, name="."):
        """
        Creates a new repository location representation.

        :param parent: the parent repository location of this entry. If the value is None, an absolute value should be defined for name.
        :param name: the name of the location, relative to parent or an absolute location, if no parent is defined.
        """
        if parent is not None and not isinstance(parent, RepositoryLocation):
            raise ValueError("'parent' must be a 'rapidminer.RepositoryLocation' object.")
        if name is not None and not isinstance(name, str):
            raise ValueError("'name' must be a string.")
        self.parent = parent
        if name is not None:
            self.name = name
        else:
            self.name = "."

    def __append_with_sep(self, path, relative_path):
        result = path
        if not result.endswith(self.__SEP__):
            result = result + self.__SEP__
        return result + relative_path

    def to_string(self, with_prefix=True):
        if self.parent is not None:
            return self.__append_with_sep(self.parent.to_string(with_prefix=with_prefix), self.name)
        elif with_prefix:
            return "repositorylocation:" + self.name
        else:
            return self.name

class ProjectLocation(Resource):
    
    def __init__(self, project, path):
        """
        Creates a new project location representation.

        :param project: the project base path. It can be an absolute path, or a relative path from the current folder.
        :param path: the relative path inside the project.
        """
        if not project or not path:
            raise ValueError("Both 'project' and 'path' arguments must not be empty.")
        self.project = project
        self.path = path
    
    def to_string(self, with_prefix=True):
        if with_prefix:
            return "git://" + self.project + ".git/" + self.path
        else:
            return self.project + "/" + self.path
    



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import os
import sys

if sys.version_info.major > 2:
    def __open__(file, mode):
        return open(file, mode, encoding="utf-8")
else:
    raise GeneralException("Python 2 is not supported. Use Python 3.")

__STDOUT_ENCODING__ = os.getenv("PYTHONIOENCODING") # encoding in python should be the same as in Studio (with -Dfile.encoding param or system global), in order to handle special characters well
if __STDOUT_ENCODING__ is None or __STDOUT_ENCODING__ == "":
    __STDOUT_ENCODING__ = "utf-8"

class GeneralException(Exception):
    """
    General exception class to errors related to the rapidminer package.
    """
    def __init__(self, msg=""):
        super(GeneralException, self).__init__(msg)

class ServerException(Exception):
    def __init__(self, msg=""):
        super(ServerException, self).__init__(msg)

class ProjectException(Exception):
    def __init__(self, msg=""):
        super(ProjectException, self).__init__(msg)

class TooManyBinomialValuesError(ProjectException):
    def __init__(self, msg=""):
        super(TooManyBinomialValuesError, self).__init__(msg)

class ValueConversionError(ProjectException):
    def __init__(self, msg=""):
        super(ValueConversionError, self).__init__(msg)

class VersionException(Exception):
    def __init__(self, product, upgrade_to):
        super(VersionException, self).__init__("You are using an older version of Python Scripting Extension in "
                                        + product + ". Upgrade to "
                                        + upgrade_to + " version to use this version of the package.")

def extract_json(res):
    """
    Returns the JSON contained in the response object, or an empty JSON.

    If the JSON contains an 'error' attribute, an exception is raised with the extracted error message.
    """
    response = {}
    if hasattr(res, 'content') and len(res.text.strip()) > 0:
        try:
            response = res.json()
        except:
            return {}
        if "error" in response:
            str = ""
            error = response["error"]
            if "type" in error:
                str += error["type"] + ": "
            if "message" in error:
                str += error["message"]
            if str == "":
                str = "Unkown error: " + str(response)
            raise ServerException(str)
    return response


class Version:
    def __init__(self, version):
        v = version.split(".")
        # cut -BETA, -SNAPSHOT from last component, etc.
        [self.major, self.minor, self.patch] = [int(v[0]), int(v[1]), int(v[2].rsplit('-', 1)[0])]
 
    def is_at_least(self, other):
        selflist = [self.major, self.minor, self.patch]
        otherlist = [other.major, other.minor, other.patch]
        for i in range(len(selflist)):
            if selflist[i] > otherlist[i]:
                return True
            elif selflist[i] < otherlist[i]:
                return False
        return True

def _is_docker_based_deployment():
    return all([var in os.environ for var in ["JUPYTERHUB_API_TOKEN", "JUPYTERHUB_API_URL", "JUPYTERHUB_USER"]])



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
from setuptools import setup, find_packages
import os
import codecs
import re

requirements = ["pandas>=1", "requests", "numpy", "zeep", "h5py", "tink"]
name = "rapidminer"
here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

version = find_version(name, "__init__.py")
readme = ""
try:
    readme = read("..", "..", "..", "..", "README.md")
except FileNotFoundError:
    try:
        # different repositories may have at a different path
        readme = read("README.md")
    except FileNotFoundError:
        # ignore
        pass

# Replaces links to other markdown files with github links to make them work on PyPi
github_baseurl = "https://github.com/rapidminer/python-rapidminer/blob/" + version + "/"
for cname in "Studio.md", "Server.md", "Scoring.md", "Project.md", "Connections.md":
    readme = readme.replace("docs/api/" + cname, github_baseurl + "docs/api/" + cname)
for pyname in "studio_examples.ipynb", "server_examples.ipynb", "project_examples.ipynb":
    readme = readme.replace("examples/" + pyname, github_baseurl + "examples/" + pyname)
readme = readme.replace("CHANGES.md", github_baseurl + "CHANGES.md")
    
setup(name=name,
      version=version,
      description='Tools for running RapidMiner processes and managing RapidMiner repositories.',
      url='https://github.com/rapidminer/python-rapidminer',
      author='RapidMiner',
      license='AGPL',
      long_description=readme,
      long_description_content_type="text/markdown",
      packages=find_packages(),
      zip_safe=False,
      install_requires=requirements,
      python_requires='>=3')



#
# This file is part of the RapidMiner Python package.
#
# Copyright (C) 2018-2021 RapidMiner GmbH
#
# This program is free software: you can redistribute it and/or modify it under the terms of the
# GNU Affero General Public License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with this program.
# If not, see https://www.gnu.org/licenses/.
#
import pandas as pd
import requests
import json
from .utilities import ServerException
from .utilities import extract_json

class Scoring:
    """
    Class that allows you to use the Real-Time Scoring agent directly on a dataset.
    """

    def __init__(self, hostname, endpoint):
        """
        Arguments:
        :param hostname: Server url (together with the port)
        :param endpoint: scoring service endpoint to use
        """
        self.url = hostname + "/services/" + endpoint

    def predict(self, dataframe):
        """
        Calls the Real-Time Scoring agent on the specified dataset and returns the result.

        Arguments:
        :param dataframe: the pandas DataFrame.
        :return: the result as a pandas DataFrame.
        """
        df_json = dataframe.to_json(orient="table")

        headers = { 'Content-type': 'application/json' }
        r = requests.post(self.url, data=df_json, headers=headers)
        response = extract_json(r)
        if r.status_code != 200:
            message = "Could not score data, status: " + str(r.status_code)
            try:
                message += ". Message: " + response["message"]
            except KeyError:
                message += ""
            raise ServerException(message)

        json_string = json.dumps(response["data"])
        df_out = pd.read_json(json_string)

        return df_out
